'''
Test class ConnectNCBI
'''
from tests.helper import *
from src.bioomics import NCBI

@ddt
class TestNCBI(TestCase):

    @skip
    @data(
        ['vertebrate_mammalian', {'vertebrate_mammalian': os.path.join(DIR_DATA, \
            'NCBI/assembly_summary/vertebrate_mammalian/assembly_summary.txt')}],
        ['wrong', {}],
    )
    @unpack
    def test_download_assembly_summary(self, input, expect):
        _, res = NCBI(DIR_DATA).download_assembly_summary([input,])
        assert res == expect

    @skip
    @data(
        ['https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/001/405/GCF_000001405.40_GRCh38.p14',
            'Homo sapiens', 'GRCh38.p14', os.path.join(DIR_DATA, 'NCBI', 'genome', \
            'Homo sapiens', 'GRCh38.p14'), 16],
    )
    @unpack
    def test_download_genome(self, ftp_path, specie, version, expect_path, expect_files):
        local_path, local_files = NCBI(DIR_DATA, False).download_genome(ftp_path, specie, version)
        assert local_path == expect_path
        assert len(local_files) == expect_files

    # TODO: confirm output
    @skip
    def test_download_gene_data(self):
        local_files = NCBI(DIR_DATA).download_gene_data()
        print(local_files)
    
    # TODO: confirm output
    @skip
    def test_download_pubmed(self):
        res = NCBI(DIR_DATA).download_pubmed()
        print(res)





"""
connect FTP using object methods
"""
from biosequtils import Dir
from ftplib import FTP
import os
import re
from sh import gunzip



class ConnFTP:
    def __init__(self, url:str, overwrite:bool=None):
        self.url = url
        self.overwrite = True if overwrite else False

    def connect(self, endpoint:str=None):
        ftp = FTP(self.url)
        ftp.login()
        try: 
            if endpoint:
                ftp.cwd(endpoint)
            return ftp
        except Exception as e:
            pass
        return None

    def list_files(self, endpoint:str=None, match:str=None):
        '''
        list ftp path of files
        not recursive
        '''
        # connect FTP
        ftp = FTP(self.url)
        ftp.login()
        if endpoint:
            ftp.cwd(endpoint)

        # scan
        ftp_files = []
        for file_name in ftp.nlst():
            if not self._is_dir(ftp, file_name):
                if match is None or re.search(match, file_name):
                    ftp_files.append((endpoint, file_name))
        return ftp_files

    def _is_dir(self, ftp, name=str):
        origin_dir = ftp.pwd()
        try:
            ftp.cwd(name)
            ftp.cwd(origin_dir)
            return True
        except Exception as e:
            pass
        return False

    def download_file(self,
            endpoint:str,
            file_name:str,
            local_path:str,
            run_gunzip:bool=None,
        ):
        '''
        download one file from FTP
        one download one connection avoiding timeout
        '''
        if run_gunzip is None: run_gunzip = True
        Dir(local_path).init_dir()
        local_file = os.path.join(local_path, file_name)
        unzip_file = local_file.replace('.gz', '')
        # doesn't download if file exists and overwrite is False
        if os.path.isfile(unzip_file) and self.overwrite is False:
            return unzip_file
        
        # connect FTP
        ftp = FTP(self.url)
        ftp.login()
        if endpoint:
            ftp.cwd(endpoint)
        ftp_file = f"{self.url}/{endpoint}/{file_name}"
        
        # download
        try:
            with open(local_file, 'wb') as f:
                ftp.retrbinary(f"RETR {file_name}", f.write)
                print(f"Download {ftp_file}")
            # unzip .gz file
            if run_gunzip and local_file.endswith('gz'):
                print(f"decompress {local_file} to {unzip_file}")
                gunzip(local_file, '-f')
                return unzip_file
            return local_file
        except Exception as e:
            print('Failure: download data from FTP, error=', e)
            os.remove(local_file)
        return None
    
    def download_files(self,
            endpoint:str=None,
            match:str=None,
            local_path:str=None,
        ):
        '''
        download files from FTP path
        That isnot recursive
        '''
        # list all files
        ftp_files = self.list_files(endpoint, match)

        # download files
        local_files = []
        for current_endpoint, file_name in ftp_files:
            local_file = self.download_file(
                endpoint = current_endpoint,
                file_name = file_name,
                local_path = local_path,
            )
            if local_file and local_file not in local_files:
                local_files.append(local_file)
        return local_files

    # TODO: need more testing
    def download_tree(self,
            local_path:str,
            endpoint:str=None,
            match:str=None
        ):
        '''
        arg: local_name is determined by os.path.join()
        Download FTP directory recursively
        '''
        # initialize endpoint
        ftp = FTP(self.url)
        ftp.login()
        origin_endpoint = ftp.pwd()

        # scan ftp path
        local_files = []
        pool = [(endpoint, local_path)]
        while pool:
            ftp.cwd(origin_endpoint)
            _endpoint, _local_path = pool.pop(0)
            if _endpoint:
                ftp.cwd(_endpoint)
            # print('FTP Path: ', ftp.pwd(), ftp.nlst())
            # check if name is directory
            for name in ftp.nlst():
                if self._is_dir(ftp, name):
                    sub_endpoint = f"{_endpoint}/{name}"
                    sub_local_path = os.path.join(_local_path, name)
                    Dir(sub_local_path).init_dir()
                    pool.append((sub_endpoint, sub_local_path))
            #download files
            local_files += self.download_files(None, match, _local_path)
        return local_files




"""
download data from NCIB FTP
"""
import os

from .connector.conn_ftp import ConnFTP
from .connector.conn_ftplib import ConnFTPlib

# lock the scope of groups in local version
ANATOMY_GROUPS = ['archaea', 'bacteria', 'fungi', 'invertebrate', 'plant',
    'protozoa', 'vertebrate_mammalian', 'vertebrate_other', 'viral', ]


class NCBI(ConnFTP):
    url = 'ftp.ncbi.nlm.nih.gov'

    def __init__(self, local_dir:str, overwrite:bool=None):
        super().__init__(url=self.url, overwrite=overwrite)
        self.local_dir = os.path.join(local_dir, "NCBI")

    def download_assembly_summary(self, groups:list=None):
        '''
        download assembly_summary.txt. That is genome metadata
        '''
        if not groups:
            groups = ANATOMY_GROUPS
        groups = [i for i in groups if i in ANATOMY_GROUPS]
        
        res = {}
        for antonomy in groups:
            outdir = os.path.join(self.local_dir, 'assembly_summary', antonomy)
            local_file = self.download_file(
                endpoint = f'genomes/refseq/{antonomy}/',
                file_name = 'assembly_summary.txt',
                local_path = outdir
            )
            res[antonomy] = local_file
        return self.local_dir, res

    def download_genome(self, ftp_path:str, specie:str, version:str=None):
        '''
        download genome including subdirectories and files
        '''
        local_path = os.path.join(self.local_dir, 'genome', specie)
        if version:
            local_path = os.path.join(local_path, version)

        # download sequences and annotations
        local_files = self.download_files(
            endpoint = ftp_path.replace('https://ftp.ncbi.nlm.nih.gov/', ''),
            match = '.gz',
            local_path = local_path,
        )
        return local_path, local_files
    
    def download_gene_data(self):
        '''
        download /gene/DATA including subdirectories and files
        '''
        local_files = self.download_tree(
            local_path = os.path.join(self.local_dir, 'gene', 'DATA'),
            endpoint = 'gene/DATA',
            match = '.gz$'
        )
        return local_files

    def download_pubmed(self):
        '''
        download /PubMed including subdirectories and files
        '''
        res = ConnFTPlib.download_tree(
            ftp_endpoint = self.url,
            ftp_path = '/pubmed',
            match = '.gz',
            local_path = self.local_dir
        )
        return res


'''
Test class 
'''
from .helper import *
from src.bioomics import ConnFTPlib as cf

@ddt
class TestConnFTPlib(TestCase):

    @skip
    @data(
        ['ftp.ncbi.nlm.nih.gov', '/pubmed', None],
        ['ftp.ncbi.nlm.nih.gov', 'pubmed', None],
        # wrong ftp path
        ['ftp.ncbi.nlm.nih.gov', 'wrong_path', 550],
        # wrong ftp url
        ['wrong_ftp', 'pubmed', 11001],
    )
    @unpack
    def test_connect_ftp(self, endpoint, ftp_path, expect):
        res = cf.connect_ftp(endpoint, ftp_path)
        assert res.get('errno') == expect

    @skip
    @data(
        ['ftp.ncbi.nlm.nih.gov', '/pubmed'],
    )
    @unpack
    def test_list_contents(self, endpoint, ftp_path):
        cf.list_contents(endpoint, ftp_path)

    @skip
    @data(
        ['ftp.ncbi.nlm.nih.gov', '/pubmed', 'gz'],
    )
    @unpack
    def test_scan_tree(self, endpoint, ftp_path, pattern):
        res = cf.scan_tree(endpoint, ftp_path, pattern)

    @data(
        ['ftp.ncbi.nlm.nih.gov', '/pubmed', 'gz'],
    )
    @unpack
    def test_download_tree(self, endpoint, ftp_path, pattern):
        local_path = os.path.join(env['DIR_DOWNLOAD'], 'NCBI')
        cf.download_tree(endpoint, ftp_path, pattern, local_path)


    @skip
    @data(
        ['ftp.ncbi.nlm.nih.gov', '/pubmed',
            {'/pubmed/baseline', '/pubmed/pubmedcommons', '/pubmed/updatefiles'},
            {'deleted.pmids.gz', 'id_list.txt.gz', 'J_Entrez.gz', 'J_Medline.gz'},
        ],
    )
    @unpack
    @mock.patch.dict(os.environ, env)
    def test_parse_contents(self, endpoint, ftp_path, expect_dirs, expect_files):
        res = cf.parse_contents(endpoint, ftp_path, 'gz')
        assert set(res['dirs']) == expect_dirs
        assert res['current_path'] == ftp_path
        assert set(res['files']) == expect_files


    @skip
    @data(
        ['/pubmed/updatefiles', 'pubmed23n1237.xml.gz', True],
        ['/pubmed/updatefiles', 'wrong_file_name', False],
    )
    @unpack
    @mock.patch.dict(os.environ, env)
    def test_download_file(self, ftp_path, file_name, expect):
        endpoint = 'ftp.ncbi.nlm.nih.gov'
        res = cf.download_file(endpoint, ftp_path, file_name, DIR_DATA)
        assert res == expect

    @skip
    @data(
        ['/pubmed/baseline', 'gz', False],
        ['/pubmed', 'gz', False],
        ['/wrong_path', 'gz', True],
        ['/pubmed', 'unmatched', True],
    )
    @unpack
    @mock.patch.dict(os.environ, env)
    def test_retrieve_file_names(self, path, pattern, expect):
        endpoint = 'ftp.ncbi.nlm.nih.gov'
        res = cf.retrieve_file_names(endpoint, path, pattern)
        assert (len(res)==0) == expect



"""
FTP of ExPASy: ftp.expasy.org
"""
import os
from .connector.conn_ftp import ConnFTP

class Expasy(ConnFTP):
    url = "ftp.expasy.org"

    def __init__(self, local_dir:str, overwrite:bool=None):
        super().__init__(url=self.url, overwrite=overwrite)
        self.local_dir = os.path.join(local_dir, 'ExPASy')
    
    def download_data(self):
        '''
        download data including subdirectories and files
        '''
        local_files = self.download_tree(
            local_path = self.local_dir,
            endpoint = 'databases',
        )
        return local_files



"""
cache data
"""
import re
import redis
import json
from biosequtils import KeyValue

# default database is 0
redis_db = {
    # NCBI
    'geneid':1,
    'gi':2,
    'ncbi_acc':3,

    # UniProt
    'uniprot_acc': 6,

}
class ConnRedis(object):
    def __init__(self, db:str=None, host:str=None, port:int=None):
        if db is None: db = ''
        self.db = redis_db.get(db, 0)
        self.host = host if host else 'localhost'
        self.port = port if port else 6379
    
        # default redis database of ontology is 0
        self.conn = redis.StrictRedis(
            host=self.host,
            port=self.port,
            db=self.db,
            decode_responses=True
         )
    
    def is_connected(self):
        """
        Test redis connection
        """
        try:
            self.conn.set('a', 1)
            self.conn.get('a')
            return True
        except Exception as e:
            print(e)
        return False

    def get_data(self, key:str):
        data = self.conn.get(key)
        if data is None:
            return None
        try:
            return json.loads(data)
        except ValueError as e:
            print(e)
            return {'data_str': data}

    def set_data(self, key:str, val):
        try:
            val = json.dumps(val)
            self.conn.set(key, val)
            return True
        except ValueError as e:
            self.conn.set(key, str(val))
            return True
        except Exception as e:
            return False
        
    def put_data(self, key:str, val):
        old = self.get_data(key)
        insert = KeyValue.merge_dict(old, val) if old else val
        self.set_data(key, insert)
    
    def get_geneid(self, geneid:str):
        '''
        Entrez GeneID: integers
        '''
        integers = re.findall(r'\d+', str(geneid))
        if integers:
            return self.get_data(integers[0])
        return None
    
    def set_geneid(self, geneid:str, data:dict):
        integers = re.findall(r'\d+', str(geneid))
        if integers:
            return self.set_data(integers[0], data)
        return False
        
    def put_geneid(self, geneid:str, data:dict):
        integers = re.findall(r'\d+', str(geneid))
        if integers:
            return self.put_data(integers[0], data)
        return False

    def get_gi(self, gi:str):
        '''
        NCBI GI(GenInfo Identifier): 8 integers
        '''
        integers = re.findall(r'\d+', str(gi))
        if integers:
            return self.get_data(integers[0])
        return None

    def set_gi(self, gi:str, data:dict):
        integers = re.findall(r'\d+', str(gi))
        if integers:
            return self.set_data(integers[0], data)
        return False
        
    def put_gi(self, gi:str, data:dict):
        integers = re.findall(r'\d+', str(gi))
        if integers:
            return self.put_data(integers[0], data)
        return False
      
    def get_uniprotkb_acc(self, acc:str):
        '''
        UniProtKB accession: 6 or 10 alphanumerical characters
        '''
        if len(acc) in (6,10):
            return self.get_data(acc)
        return None

    def set_uniprotkb_acc(self, acc:str, data:dict):
        if len(acc) in (6,10):
            return self.set_data(acc, data)
        return False
        
    def put_uniprotkb_acc(self, acc:str, data:dict):
        if len(acc) in (6,10):
            return self.put_data(acc, data)
        return False

    def get_ncbi_acc(self, acc:str):
        '''
        NCBI sequence accession: xx_xxxxxxx.x
        <two leters>_<6 digits>.<version>
        '''
        return self.get_data(acc)

    def set_ncbi_acc(self, acc:str, data:dict):
        return self.set_data(acc, data)
        
    def put_ncbi_acc(self, acc:str, data:dict):
        return self.put_data(acc, data)



# connector
from .connector.conn_ftp import ConnFTP
from .connector.conn_ftplib import ConnFTPlib
from .connector.conn_redis import ConnRedis

# database
from .ncbi import NCBI, ANATOMY_GROUPS
from .rnacentral import RNACentral
from .mirbase import Mirbase
from .expasy import Expasy




'''
'''

from .helper import *
from src.bioomics import RNACentral

@ddt
class TestRNACentral(TestCase):

    @data(
        ['lncbook.fasta', os.path.join(DIR_DATA, 'RNACentral', 'lncbook.fasta')],
        ['pirbase.fasta', os.path.join(DIR_DATA, 'RNACentral', 'pirbase.fasta')],
        ['wrong', None],
    )
    @unpack
    def test_download_sequence(self, file_name, expect):
        local_path, local_file = RNACentral(DIR_DATA, False).download_sequence(file_name)
        assert local_path == os.path.join(DIR_DATA, 'RNACentral')
        assert local_file == expect


'''
Test class 
'''
from .helper import *
from src.bioomics import ConnRedis

@ddt
class TestConnRedis(TestCase):

    def setUp(self):
        self.c = ConnRedis()
        self.c.conn.set('a', 1)
        self.c.conn.set('b', json.dumps({'c':[1,2], 'd':{'e':'f'}}))

    def test_connection(self):
        self.c.is_connected()

    @data(
        ['a', 1],
        ['wrong', None],
        ['b', {'c':[1,2], 'd':{'e':'f'}}],
    )
    @unpack
    def test_get_data(self, key, expect):
        res = self.c.get_data(key)
        assert res == expect

    @data(
        ['a', 1],
        ['b', {'c':[1,2], 'd':{'e':'f'}}],
        ['c', None],
        ['d', [1,2,'a']],
    )
    @unpack
    def test_set_data(self, key, val):
        self.c.set_data(key, val)
        res = self.c.get_data(key)
        assert res == val
    
    def test_put_data(self):
        key, val = '1', {'id':'2'}
        self.c.set_data(key, val)
        res = self.c.get_data(key)
        assert res == val

        #same key but different values
        val = {'id':'3'}
        self.c.put_data(key, val)
        res = self.c.get_data(key)
        assert res == {'id':['2','3']}

        # duplicaete value
        val = {'id':'3'}
        self.c.put_data(key, val)
        res = self.c.get_data(key)
        assert res == {'id':['2','3']}

        # new key
        key, val = '2', {'id':'3'}
        self.c.set_data(key, val)
        res = self.c.get_data(key)
        assert res == val


    @data(
        ['856646', []],
        ['856646', None],
        ['856646', {'symbol': 'CAN1', 'pmid':[34,66]}],
        ['GeneID:856646', {'symbol': 'CAN1'}],
        ['GeneID: 856646', {'symbol': 'CAN1'}],
        ["geneid: 856646 \t ", {'symbol': 'CAN1'}],
    )
    @unpack
    def test_geneid(self, geneid, data):
        c = ConnectRedis('geneid')
        c.set_geneid(geneid, data)
        res = c.get_geneid(geneid)
        assert res == data

    @data(
        ['A2BC19', 'abc'],
        ['A2BC19', []],
        ['A2BC19', {'id': 'A2BC19_HELPX'}],
        ['A2BC19', None],
        ['wrong', None],
    )
    @unpack
    def test_uniprotkb_acc(self, acc, data):
        c = ConnectRedis('uniprot_acc')
        c.set_uniprotkb_acc(acc, data)
        res = c.get_uniprotkb_acc(acc)
        assert res == data
       
    @data(
        ['NC_000001.11', 'abc'],
        ['NC_000001.11', {'gi': 568815597}],
    )
    @unpack
    def test_ncbi_acc(self, acc, data):
        c = ConnectRedis('ncbi_acc')
        c.set_ncbi_acc(acc, data)
        res = c.get_ncbi_acc(acc)
        assert res == data


'''
Test class 
'''
from tests.helper import *

from src.bioomics import ConnFTP

@ddt
class TestConnFTP(TestCase):

    def setUp(self):
        self.c = ConnFTP('ftp.ncbi.nlm.nih.gov')

    @data(
        ['blast', True],
        ['robts.txt', False]
    )
    @unpack
    def test_is_dir(self, name, expect):
        ftp = self.c.connect()
        res = self.c._is_dir(ftp, name)

    @data(
        [
            'geo/series/GSE3nnn/GSE3341/matrix',
            [('geo/series/GSE3nnn/GSE3341/matrix', 'GSE3341_series_matrix.txt.gz')],
        ],
    )
    @unpack
    def test_list_files(self, path, expect):
        res = self.c.list_files(path)
        assert res == expect

    @data(
        ['geo/series/GSE3nnn/GSE3341/matrix', '.gz', 1],
        ['geo/series/GSE3nnn/GSE3341/', '.gz', 0],
        ['geo/series/GSE3nnn/GSE3341/', None, 0],
    )
    @unpack
    def test_download_files(self, path, pattern, expect):
        res = self.c.download_files(path, pattern, DIR_DATA)
        assert len(res) == expect


    @data(
        ['gene/tools/', 'README', None, 'README'],
        ['gene/DATA/', 'gene_orthologs.gz', False, 'gene_orthologs.gz'],
        ['gene/DATA/', 'gene_orthologs.gz', True, 'gene_orthologs'],
    )
    @unpack
    def test_download_file(self, ftp_path, file_name, run_gunzip, expect):
        res = self.c.download_file(ftp_path, file_name, DIR_DATA, run_gunzip)
        expect = os.path.join(DIR_DATA, expect)
        assert res == expect



    # @skip
    # @data(
    #     ['gene/tools/', None, 3],
    #     # ['pubmed/updatefiles/', 'gz', 0],
    # )
    # @unpack
    # def test_download_tree(self, path, pattern, expect):
    #     res = self.c.download_tree(DIR_DATA, path, pattern)
    #     assert len(res) >= expect


'''
https://rnacentral.org
'''
import os
from .connector.conn_ftp import ConnFTP


class RNACentral(ConnFTP):
    url = "ftp.ebi.ac.uk"
    endpoint = '/pub/databases/RNAcentral/current_release/'

    def __init__(self, local_dir:str, overwrite:bool=None):
        super().__init__(url=self.url, overwrite=overwrite)
        self.local_dir = os.path.join(local_dir, "RNACentral")
    
    def download_sequence(self, file_name:str) -> tuple:
        _endpoint = f'{self.endpoint}sequences/by-database'
        local_file = self.download_file(
            endpoint = _endpoint,
            file_name = file_name,
            local_path = self.local_dir,
            run_gunzip = False,
        )
        return self.local_dir, local_file


'''
MiRBase
'''
from .helper import *
from src.bioomics import Mirbase

class TestMirbase(TestCase):
    def setUp(self):
        self.c = Mirbase(DIR_DATA, False)

    def test_download_hairpin(self):
        local_path, local_file = self.c.download_hairpin()
        assert local_path == os.path.join(DIR_DATA, 'miRBase')
        assert local_file == os.path.join(DIR_DATA, 'miRBase', 'hairpin.fa')

    def test_download_mature(self):
        local_path, local_file = self.c.download_mature()
        assert local_path == os.path.join(DIR_DATA, 'miRBase')
        assert local_file == os.path.join(DIR_DATA, 'miRBase', 'mature.fa')




'''
miRBase: https://mirbase.org/
'''
import os
from Bio import SeqIO
from biosequtils import Dir
from io import StringIO
import lxml.html as html
from urllib.request import urlopen


class Mirbase:
    url = 'https://mirbase.org/download/CURRENT/'

    def __init__(self, local_dir:str, overwrite:bool=None):
        self.local_dir = os.path.join(local_dir, "miRBase")
        Dir(self.local_dir).init_dir()
        self.overwrite = True if overwrite else False
    
    def download_hairpin(self):
        local_file = os.path.join(self.local_dir, 'hairpin.fa')
        if self.overwrite or not os.path.isfile(local_file):
            self.parse_fasta(f"{self.url}/hairpin.fa", local_file)
        return self.local_dir, local_file

    def download_mature(self):
        local_file = os.path.join(self.local_dir, 'mature.fa')
        if self.overwrite or not os.path.isfile(local_file):
            self.parse_fasta(f"{self.url}/mature.fa", local_file)
        return self.local_dir, local_file

    def parse_fasta(self, url_path, local_file):
        # parse xml
        parsed = html.parse(urlopen(url_path))
        doc = parsed.getroot()
        for br in doc.xpath("*//br"):
            br.tail = "\n" + br.tail if br.tail else "\n"
        # parse fasta
        fa_io = StringIO(doc.text_content())
        records = SeqIO.parse(fa_io, 'fasta')
        with open(local_file, 'w') as f:
            for rec in records:
                SeqIO.write(rec, f, 'fasta')



'''
Test class 
'''
from .helper import *
from src.bioomics import Expasy


@ddt
class TestExpasy(TestCase):

    def setUp(self):
        self.c = Expasy(DIR_DATA, False)

    def list_files(self):
        res = self.c.list_files()
        print(res)

    def test_download_data(self):
        res = self.c.download_data()
        print(res)



"""
connect FTP using static methods
1. tremendous files have to be download
2. FTP login can't be alive long time.
3. internet connection is low-speed
"""
from biosequtils import Dir, HandleJson
import ftplib
import os
import time
from typing import Callable


class ConnFTPlib:

    @staticmethod
    def connect_ftp(
            ftp_endpoint:str,
            ftp_path:str,
            func:Callable=None,
            try_times:str=None,
            **kwargs
        ):
        '''
        connection handler
        '''
        error = ''
        if try_times is None: try_times = 3
        for i in range(1, try_times+1):
            try:
                res = None
                ftp = ftplib.FTP(ftp_endpoint)
                ftp.login()
                ftp.cwd(ftp_path)
                if func:
                    res = func(ftp, **kwargs)
                ftp.close()
                return { 'return': res }
            except ftplib.error_perm as err1:
                print(f"Failure: {ftp_endpoint} is ok. but the path {ftp_path} is wrong.")
                return {
                        'errno': 550,
                        'error': err1,
                    }
            except ftplib.all_errors as err2:
                if i == try_times:
                    print(f"Failure: can't target {ftp_endpoint}/{ftp_path}. " + 
                            f"Try {try_times} times. error={error}")
                    return {
                        'error': err2,
                    }
                time.sleep(2)
            except Exception as e:
                print(f"Error: error={e}")
                return {
                        'error': e,
                    }

    @staticmethod
    def _download(ftp, **kwargs):
        ftp_file_name = kwargs.get('ftp_file_name')
        local_file = kwargs.get('local_file')
        if ftp_file_name and local_file:
            try:
                with open(local_file, 'wb') as f:
                    ftp.retrbinary(f"RETR {ftp_file_name}", f.write)
                    print(f"Download FTP data {ftp_file_name} to {local_file}.")
                return True
            except Exception as e:
                print(f"Failure: can't download FTP data: {ftp_file_name}. error={e}")
                if os.path.isfile(local_file):
                    os.remove(local_file)
        return False
    
    @staticmethod
    def list_contents(ftp_endpoint:str, ftp_path:str):
        def _func(ftp):
            contents = []
            ftp.dir(contents.append)
            return contents
        res = ConnFTPlib.connect_ftp(ftp_endpoint, ftp_path, _func)
        if 'return' in res:
            return res['return']


    @staticmethod
    def parse_contents(ftp_endpoint:str, ftp_path:str, pattern:str=None):
        parser = {
            'dirs': [],
            'files': [],
            'current_path': ftp_path,
        }
        contents = ConnFTPlib.list_contents(ftp_endpoint, ftp_path)
        for i in contents:
            items = i.split(' ')
            if items[0].startswith('d'):
                parser['dirs'].append(f"{ftp_path}/{items[-1]}")
            elif items[0].startswith('-'):
                if pattern is None or (pattern is not None \
                        and items[-1].endswith(pattern)):
                    parser['files'].append(items[-1])
        return parser

    @staticmethod
    def scan_tree(ftp_endpoint:str, ftp_path:str, pattern:str=None)->list:
        '''
        anonymous user login FTP site
        walk through the directory recrusively
        '''
        path = []
        pool = [ftp_path,]
        while pool:
            current_ftp_path = pool.pop(0)
            list_contents = ConnFTPlib.parse_contents(
                ftp_endpoint,  current_ftp_path, pattern
            )
            pool += list_contents['dirs']
            for name in list_contents['files']:
                path.append((list_contents['current_path'], name))
        return path


    @staticmethod
    def retrieve_file_names(
            ftp_endpoint:str,
            ftp_path:str,
            name_pattern:str=None
        )->list:
        '''
        Suppose that large number of files exists in one directory.
        Don't consider sub-directory
        '''
        # scan ftp path
        res = ConnFTPlib.connect_ftp(
            ftp_endpoint,
            ftp_path,
            lambda ftp: ftp.nlst()
        )
        if 'return' in res:
            file_names = []
            for name in res['return']:
                if name_pattern is None or (name_pattern is not None\
                        and name.endswith(name_pattern)):
                    # print(ftp_path, name)
                    file_names.append(name)
            return file_names
        return []


    @staticmethod
    def download_file(
            ftp_endpoint:str,
            ftp_path:str,
            ftp_file_name:str,
            local_path:str
        )->bool:
        '''
        anonymous user login FTP site
        Download one file at a time. Try 3 times in default
        local file would be covered if that exists.
        '''
        local_file = os.path.join(local_path, ftp_file_name)
        res = ConnFTPlib.connect_ftp(
            ftp_endpoint=ftp_endpoint,
            ftp_path=ftp_path,
            func=ConnFTPlib._download,
            ftp_file_name=ftp_file_name,
            local_file=local_file,
        )
        if 'return' in res:
            return res['return']
        return False
    

    @staticmethod
    def download_tree(
            ftp_endpoint:str,
            ftp_path:str,
            match:str,
            local_path:str
        )->bool:
        '''
        anonymous user login FTP site
        Download a large number of files, but Download one file at a time. 
        local file would be covered if that exists.
        '''
        tree_path = ConnFTPlib.scan_tree(ftp_endpoint, ftp_path, match)
        for current_ftp_path, file_name in tree_path:
            if current_ftp_path.startswith('/'):
                current_ftp_path = current_ftp_path[1:]
            dir_names = [local_path,] + current_ftp_path.split('/')
            current_local_path = os.path.join(*dir_names)
            Dir(current_local_path).init_dir()
            # print(current_ftp_path, current_local_path, file_name)
            # download file
            ConnFTPlib.download_file(
                ftp_endpoint,
                current_ftp_path,
                file_name,
                current_local_path
            )


